@@ -1,6 +1,6 @@
-// $Header: /cvsroot/libcwd/libcwd/bfd.cc,v 1.164 2004/05/27 03:03:43 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/bfd.cc,v 1.173 2004/10/07 18:00:16 libcw Exp $
 //
-// Copyright (C) 2000 - 2003, by
+// Copyright (C) 2000 - 2004, by
 // 
 // Carlo Wood, Run on IRC <carlo@alinoe.com>
 // RSA-1024 0x624ACAD5 1997-01-26                    Sign & Encrypt
@@ -86,8 +86,8 @@ extern char** environ;
 namespace libcwd {
   namespace _private_ {
     extern void demangle_symbol(char const* in, _private_::internal_string& out);
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
-    extern __gnu_cxx::_STL_mutex_lock* _ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr;
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+    extern char* pool_allocator_lock_symbol_ptr;
 #endif
   } // namespace _private_
 
@@ -279,10 +279,10 @@ void bfd_close(bfd* abfd)
 #endif
       }
 
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
-      void bfile_ct::initialize(char const* filename, void* base, bool is_libstdcpp LIBCWD_COMMA_TSD_PARAM)
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+      void bfile_ct::initialize(char const* filename, void* base LIBCWD_COMMA_ALLOC_OPT(bool is_libc), bool is_libstdcpp LIBCWD_COMMA_TSD_PARAM)
 #else
-      void bfile_ct::initialize(char const* filename, void* base LIBCWD_COMMA_TSD_PARAM)
+      void bfile_ct::initialize(char const* filename, void* base LIBCWD_COMMA_ALLOC_OPT(bool is_libc) LIBCWD_COMMA_TSD_PARAM)
 #endif
       {
 #if CWDEBUG_DEBUGM
@@ -362,8 +362,11 @@ void bfd_close(bfd* abfd)
 
 	if (M_number_of_symbols > 0)
 	{
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
-	  Elf32_Off S_lock_value;
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+	  Elf32_Off S_lock_value = 0;
+#endif
+#if CWDEBUG_ALLOC
+	  Elf32_Off exit_funcs = 0;
 #endif
 #if !CWDEBUG_LIBBFD
 	  size_t s_end_offset = M_abfd->M_s_end_offset;
@@ -567,15 +570,21 @@ void bfd_close(bfd* abfd)
 	  asymbol** se2 = &M_symbol_table[M_number_of_symbols - 1];
 	  for (asymbol** s = M_symbol_table; s <= se2;)
 	  {
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
 	    if (is_libstdcpp && strcmp((*s)->name,
-#if __GNUC_MINOR__ == 4
+#if __GNUC_PATCHLEVEL__ == 0
 	    "_ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE"
+#elif __GNUC_PATCHLEVEL__ == 1
+	    "_ZN9__gnu_cxx11__pool_baseILb1EE7_S_lockE"
 #else
-	    "_ZN9__gnu_cxx12__pool_allocIcE7_S_lockE"
+	    "_ZN14__gnu_internal17palloc_init_mutexE"
 #endif
 	    ) == 0)
 	      S_lock_value = bfd_get_section(*s)->vma + (*s)->value;
+#endif // LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+#if CWDEBUG_ALLOC
+	    if (is_libc && strcmp((*s)->name, "__exit_funcs") == 0)
+	      exit_funcs = bfd_get_section(*s)->vma + (*s)->value;
 #endif
 	    if (((*s)->name[0] == '.' && (*s)->name[1] == 'L')
 	        || ((*s)->flags & (BSF_SECTION_SYM|BSF_OLD_COMMON|BSF_NOT_AT_END|BSF_INDIRECT|BSF_DYNAMIC)) != 0
@@ -594,11 +603,20 @@ void bfd_close(bfd* abfd)
 
 	  if (M_function_symbols.size() > 0)
 	  {
-	    asymbol const* last_symbol = (*M_function_symbols.rbegin()).get_symbol();
+	    // M_function_symbols is a set<> that sorts in reverse order (using symbol_key_greater)!
+	    // So use begin() here in order to get the last symbol.
+	    asymbol const* last_symbol = (*M_function_symbols.begin()).get_symbol();
 	    if (symbol_size(last_symbol) == 100000)
 	    {
 	      BFD_RELEASE_WRITE_LOCK;
-	      Dout( dc::warning, "Unknown size of symbol " << last_symbol->name);
+#if CWDEBUG_ALLOC
+	      int saved_internal = __libcwd_tsd.internal;
+	      __libcwd_tsd.internal = 0;
+#endif
+	      Dout(dc::warning, "Unknown size of symbol " << last_symbol->name);
+#if CWDEBUG_ALLOC
+	      __libcwd_tsd.internal = saved_internal;
+#endif
 	      BFD_ACQUIRE_WRITE_LOCK;
 	    }
 	    if (!M_size)
@@ -625,10 +643,13 @@ void bfd_close(bfd* abfd)
 	    Debug( libcw_do.restore(state) );
 	  }
 #endif
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
 	  if (is_libstdcpp && S_lock_value)
-	    _private_::_ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr =
-	        (__gnu_cxx::_STL_mutex_lock*)((char*)M_lbase + S_lock_value);
+	    _private_::pool_allocator_lock_symbol_ptr = (char*)M_lbase + S_lock_value;
+#endif
+#if CWDEBUG_ALLOC
+          if (is_libc && exit_funcs)
+	    _private_::__exit_funcs_ptr = (_private_::exit_function_list**)((char*)M_lbase + exit_funcs);
 #endif
 	}
 
@@ -1006,22 +1027,25 @@ void bfd_close(bfd* abfd)
 	else
 	  Dout(dc::bfd|continued_cf|flush_cf, "Loading debug symbols from " << name << " (" << l_addr << ") ... ");
 	bfile_ct* object_file;
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
-	bool is_libstdcpp;
 	char const* slash = strrchr(name, '/');
 	if (!slash)
 	  slash = name - 1;
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+	bool is_libstdcpp;
 	is_libstdcpp = (strncmp("libstdc++.so", slash + 1, 12) == 0);
+#endif
+#if CWDEBUG_ALLOC
+	bool is_libc = (strncmp("libc.so", slash + 1, 7) == 0);
 #endif
 	LIBCWD_DEFER_CANCEL;
 	BFD_ACQUIRE_WRITE_LOCK;
 	set_alloc_checking_off(LIBCWD_TSD);
 	object_file = new bfile_ct(name, l_addr);
 	BFD_RELEASE_WRITE_LOCK;
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ >= 4
-	object_file->initialize(name, l_addr, is_libstdcpp LIBCWD_COMMA_TSD);
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
+	object_file->initialize(name, l_addr LIBCWD_COMMA_ALLOC_OPT(is_libc), is_libstdcpp LIBCWD_COMMA_TSD);
 #else
-	object_file->initialize(name, l_addr LIBCWD_COMMA_TSD);
+	object_file->initialize(name, l_addr LIBCWD_COMMA_ALLOC_OPT(is_libc) LIBCWD_COMMA_TSD);
 #endif
 	set_alloc_checking_on(LIBCWD_TSD);
 	LIBCWD_RESTORE_CANCEL;
@@ -1165,10 +1189,13 @@ void bfd_close(bfd* abfd)
 	  }
 #endif
 #endif
-	    if (l->l_addr)
-	      load_object_file(l->l_name, reinterpret_cast<void*>(l->l_addr));
-	    else if (l->l_name && (l->l_name[0] == '/') || (l->l_name[0] == '.'))
-	      load_object_file(l->l_name, unknown_l_addr);
+	    if (l->l_name && (l->l_name[0] == '/') || (l->l_name[0] == '.'))
+	    {
+	      if (l->l_addr)
+		load_object_file(l->l_name, reinterpret_cast<void*>(l->l_addr));
+	      else
+		load_object_file(l->l_name, unknown_l_addr);
+	    }
 	  }
 
 	  LIBCWD_DEFER_CANCEL;
@@ -1327,7 +1354,7 @@ typedef location_ct bfd_location_ct;
 	// MT: `WST_initialized' is only false when we're still Single Threaded.
 	//     Therefore it is safe to call ST_* functions.
 
-#if CWDEBUG_ALLOC && __GNUC_MINOR__ < 3
+#if CWDEBUG_ALLOC && LIBCWD_IOSBASE_INIT_ALLOCATES
         if (!_private_::WST_ios_base_initialized && _private_::inside_ios_base_Init_Init())
 	{
 	  M_object_file = NULL;
@@ -1413,6 +1440,7 @@ already_loaded:
         LIBCWD_Dout(dc::bfd, "No object file for address " << addr);
 	M_object_file = NULL;
 	M_func = unknown_function_c;
+	M_unknown_pc = addr;
 	return;
       }
       M_object_file = object_file->get_object_file();
@@ -1576,22 +1604,6 @@ already_loaded:
       return *this;
     }
 
-    /**
-     * \brief Write \a location to ostream \a os.
-     * \ingroup group_locations
-     *
-     * Write the contents of a location_ct object to an ostream in the form "source-%file:line-number",
-     * or writes "<unknown location>" when the location is unknown.
-     */
-    std::ostream& operator<<(std::ostream& os, location_ct const& location)
-    {
-      if (location.M_known)
-	os << location.M_filename << ':' << location.M_line;
-      else
-	os << "<unknown location>";
-      return os;
-    }
-
 } // namespace libcwd
 
 #if defined(LIBCWD_DLOPEN_DEFINED) && defined(HAVE_DLOPEN)
