@@ -1,6 +1,6 @@
-// $Header: /cvsroot/libcwd/libcwd/debug.cc,v 1.130 2004/05/27 03:03:43 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/debug.cc,v 1.143 2004/09/26 15:50:58 libcw Exp $
 //
-// Copyright (C) 2000 - 2003, by
+// Copyright (C) 2000 - 2004, by
 // 
 // Carlo Wood, Run on IRC <carlo@alinoe.com>
 // RSA-1024 0x624ACAD5 1997-01-26                    Sign & Encrypt
@@ -14,7 +14,7 @@
 #include "sys.h"
 #include <libcwd/config.h>
 
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
 // This has to be very early (must not have been included elsewhere already).
 #define private public  // Ugly, I know.
 #if (__GNUC__ == 3 && __GNUC_MINOR__ < 4)
@@ -29,8 +29,10 @@ extern "C" char* getenv(char const* name);	// Needed before including ext/pool_a
 #include <cerrno>
 #include <iostream>
 #include <algorithm>
+#ifdef HAVE_SYS_RESOURCE_H
 #include <sys/time.h>     	// Needed for setrlimit()
 #include <sys/resource.h>	// Needed for setrlimit()
+#endif
 #include <cstdlib>		// Needed for Exit() (C99)
 #include <new>
 #include "cwd_debug.h"
@@ -38,15 +40,15 @@ extern "C" char* getenv(char const* name);	// Needed before including ext/pool_a
 #include <libcwd/private_internal_stringbuf.h>
 #include <libcwd/private_bufferstream.h>
 #include "private_debug_stack.inl"
+#include <libcwd/class_location.inl>
 
 extern "C" int raise(int);
+extern "C" void _exit(int);
 
 #if LIBCWD_THREAD_SAFE
 using libcwd::_private_::rwlock_tct;
-#if __GNUC_MINOR__ != 5
 using libcwd::_private_::debug_objects_instance;
 using libcwd::_private_::debug_channels_instance;
-#endif
 #define DEBUG_OBJECTS_ACQUIRE_WRITE_LOCK	rwlock_tct<libcwd::_private_::debug_objects_instance>::wrlock()
 #define DEBUG_OBJECTS_RELEASE_WRITE_LOCK	rwlock_tct<libcwd::_private_::debug_objects_instance>::wrunlock()
 #define DEBUG_OBJECTS_ACQUIRE_READ_LOCK		rwlock_tct<libcwd::_private_::debug_objects_instance>::rdlock()
@@ -77,65 +79,86 @@ using libcwd::_private_::debug_channels_instance;
 #endif // !LIBCWD_THREAD_SAFE
 
 #define NEED_SUPRESSION_OF_MALLOC_AND_BFD (__GNUC__ == 3 && \
-    ((__GNUC_MINOR__ == 2 && __VERSION__ [3] == 0) || \
-     (__GNUC_MINOR__ == 1 && __VERSION__ [4] == '1') || \
+    ((__GNUC_MINOR__ == 2 && __GNUC_PATCHLEVEL__ == 0) || \
+     (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ == 1) || \
      (__GNUC_MINOR__ == 0)))
 
 namespace libcwd {
 
 namespace _private_ {
 
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
-#if (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
-__gnu_cxx::_STL_mutex_lock* _ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr;
+// gcc 4.0 use mt_allocator.h as default, which doesn't use locks at all.
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
+
+#if (__GNUC_MINOR__ == 4)
+#if (__GNUC_PATCHLEVEL__ <= 1)
+__gnu_cxx::_STL_mutex_lock* pool_allocator_lock_symbol_ptr;
+#else
+__gthread_mutex_t* pool_allocator_lock_symbol_ptr;
+#endif
 #endif
 
 // The following tries to take the "node allocator" lock -- the lock of the
 // default allocator for threaded applications.
-__inline__
+inline
 bool allocator_trylock(void)
 {
-#if (__GNUC__ < 3 || __GNUC_MINOR__ == 0)
-  if (!(__NODE_ALLOCATOR_THREADS)) return true;
-#endif
-#if (__GNUC__ == 3 && __GNUC_MINOR__ < 4)
+#if (__GNUC_MINOR__ < 4)
+
 #if !defined(__GTHREAD_MUTEX_INIT) && defined(__GTHREAD_MUTEX_INIT_FUNCTION)
   if (!std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_init_flag)
     std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_initialize();
 #endif
+
   return (__gthread_mutex_trylock(&std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_lock) == 0);
+
+#else // __GNUC_MINOR__ == 4
+
+  if (!pool_allocator_lock_symbol_ptr)
+    return false;
+
+  #if !defined(__GTHREAD_MUTEX_INIT) && defined(__GTHREAD_MUTEX_INIT_FUNCTION)
+  #if (__GNUC_PATCHLEVEL__ <= 1)
+    if (!pool_allocator_lock_symbol_ptr->_M_init_flag)
+      pool_allocator_lock_symbol_ptr->_M_initialize();
+  #else
+  // If the lock is not initialized, it should be initialized here.
+  #error "Sorry, not implemented yet."
+  #endif
+  #endif
+
+#if (__GNUC_PATCHLEVEL__ <= 1)
+  return (__gthread_mutex_trylock(&pool_allocator_lock_symbol_ptr->_M_lock) == 0);
 #else
-#if !defined(__GTHREAD_MUTEX_INIT) && defined(__GTHREAD_MUTEX_INIT_FUNCTION)
-  if (!LIBCWD_POOL_ALLOC<true, 0>::_S_lock._M_init_flag)
-    LIBCWD_POOL_ALLOC<true, 0>::_S_lock._M_initialize();
-#endif
-  return (_ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr &&
-          __gthread_mutex_trylock(&_ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr->_M_lock) == 0);
+  return (__gthread_mutex_trylock(pool_allocator_lock_symbol_ptr) == 0);
 #endif
+
+#endif // __GNUC_MINOR__ == 4
 }
 
 // The following unlocks the node allocator.
-__inline__
+inline
 void allocator_unlock(void)
 {
-#if (__GNUC__ < 3 || __GNUC_MINOR__ == 0)
-  if (!(__NODE_ALLOCATOR_THREADS)) return;
-#endif
-#if (__GNUC__ == 3 && __GNUC_MINOR__ < 4)
+#if (__GNUC_MINOR__ < 4)
 #if !defined(__GTHREAD_MUTEX_INIT) && defined(__GTHREAD_MUTEX_INIT_FUNCTION)
   if (!std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_init_flag)
     std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_initialize();
 #endif
   __gthread_mutex_unlock(&std::__default_alloc_template<true, 0>::_S_node_allocator_lock._M_lock);
+
+#else // __GNUC_MINOR__ == 4
+
+#if (__GNUC_PATCHLEVEL__ <= 1)
+  __gthread_mutex_unlock(&pool_allocator_lock_symbol_ptr->_M_lock);
 #else
-#if !defined(__GTHREAD_MUTEX_INIT) && defined(__GTHREAD_MUTEX_INIT_FUNCTION)
-  if (!LIBCWD_POOL_ALLOC<true, 0>::_S_lock._M_init_flag)
-    LIBCWD_POOL_ALLOC<true, 0>::_S_lock._M_initialize();
-#endif
-  __gthread_mutex_unlock(&_ZN9__gnu_cxx12__pool_allocILb1ELi0EE7_S_lockE_ptr->_M_lock);
+  __gthread_mutex_unlock(pool_allocator_lock_symbol_ptr);
 #endif
+
+#endif // __GNUC_MINOR__ == 4
 }
-#endif // LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+
+#endif // LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
 
 } // namespace _private_
 
@@ -196,7 +219,7 @@ void allocator_unlock(void)
       // ends_on_newline	: This output ends on a newline.
       // possible_nonewline_cf	: When `ends_on_newline' is false, then that was caused by the use of nonewline_cf.
 
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
       typedef debug_message_st* msgbuf_t;
       msgbuf_t msgbuf;
       // Queue the message when the default (STL) allocator is locked and it could be that we
@@ -225,12 +248,12 @@ void allocator_unlock(void)
 	msgbuf = (msgbuf_t)malloc(curlen + extra_size);
 	free_msgbuf = true;
       }
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
       this->sgetn(msgbuf->buf, curlen);
 #else
       this->sgetn(msgbuf, curlen);
 #endif
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
       if (queue_msg)	// Inside a call to malloc and possibly owning lock of std::LIBCWD_POOL_ALLOC<true, 0>?
       {
 	// We don't write debug output to the final ostream when inside malloc and std::LIBCWD_POOL_ALLOC<true, 0> is locked.
@@ -298,7 +321,7 @@ void allocator_unlock(void)
 	  locked_os->write("<continued> ", 12);
 	}
 #endif // LIBCWD_THREAD_SAFE
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
 	debug_message_st* message = debug_object.queue_top;
 	if (message)
 	{
@@ -363,7 +386,7 @@ void allocator_unlock(void)
 	--LIBCWD_DO_TSD_MEMBER_OFF(libcw_do);
 	_private_::set_library_call_off(saved_internal LIBCWD_COMMA_TSD);
 #endif
-#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC
+#if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3
       }
 #endif
       if (free_msgbuf)
@@ -804,7 +827,7 @@ void allocator_unlock(void)
 #if LIBCWD_THREAD_SAFE
       LIBCWD_ENABLE_CANCEL;
 #endif
-      _Exit(6);		// Never reached.
+      _exit(6);		// Never reached.
     }
 
     size_t debug_string_ct::calculate_capacity(size_t size)
@@ -934,6 +957,7 @@ void allocator_unlock(void)
       set_alloc_checking_off(LIBCWD_TSD);
       void* new_debug_string = malloc(sizeof(debug_string_stack_element_ct));
       LIBCWD_TSD_MEMBER(M_marker_stack) = new (new_debug_string) debug_string_stack_element_ct(LIBCWD_TSD_MEMBER(marker));
+      set_alloc_checking_on(LIBCWD_TSD);
       LIBCWD_TSD_MEMBER(M_marker_stack)->next = current_marker_stack;
     }
 
@@ -970,7 +994,9 @@ void allocator_unlock(void)
       if (NEED_SUPRESSION_OF_MALLOC_AND_BFD)
       {
         channels::dc::malloc.off();
+#if CWDEBUG_LOCATION
         channels::dc::bfd.off();
+#endif
       }
 
       // Skip `start()' for a `continued' debug output.
@@ -1136,7 +1162,9 @@ void allocator_unlock(void)
       if (NEED_SUPRESSION_OF_MALLOC_AND_BFD)
       {
         channels::dc::malloc.on();
+#if CWDEBUG_LOCATION
         channels::dc::bfd.on();
+#endif
       }
 
       // Skip `finish()' for a `continued' debug output.
@@ -1229,7 +1257,7 @@ void allocator_unlock(void)
 	  _private_::rwlock_tct<_private_::threadlist_instance>::rdunlock();
 	  LIBCWD_ENABLE_CANCEL;
 #endif
-	  _Exit(254);	// Exit without calling global destructors.
+	  _exit(254);	// Exit without calling global destructors.
 	}
 	if ((current->mask & wait_cf))
 	{
@@ -1303,7 +1331,7 @@ void allocator_unlock(void)
     {
       finish(debug_object, channel_set LIBCWD_COMMA_TSD);
       DoutFatal( dc::core, "Don't use `DoutFatal' together with `continued_cf', use `Dout' instead.  (This message can also occur when using DoutFatal correctly but from the constructor of a global object)." );
-      _Exit(1);	// This is never reached, but g++ 3.3.x -O2 thinks it is.
+      _exit(1);	// This is never reached, but g++ 3.3.x -O2 thinks it is.
     }
 
 #if LIBCWD_THREAD_SAFE
@@ -1524,11 +1552,13 @@ void allocator_unlock(void)
 	for(_private_::debug_channels_ct::container_type::const_iterator i(_private_::debug_channels.read_locked().begin());
 	    i != _private_::debug_channels.read_locked().end(); ++i)
 	{
-	  LibcwDoutScopeBegin(DEBUGCHANNELS, debug_object, dc::always|noprefix_cf);
+	  LibcwDoutScopeBegin(LIBCWD_DEBUGCHANNELS, debug_object, dc::always|noprefix_cf);
 	  if (NEED_SUPRESSION_OF_MALLOC_AND_BFD)
 	  {
 	    channels::dc::malloc.on();
+#if CWDEBUG_LOCATION
 	    channels::dc::bfd.on();
+#endif
 	  }
 	  LibcwDoutStream.write(LIBCWD_DO_TSD_MEMBER(debug_object, margin).c_str(), LIBCWD_DO_TSD_MEMBER(debug_object, margin).size());
 	  LibcwDoutStream.write((*i)->get_label(), WST_max_len);
@@ -1539,7 +1569,9 @@ void allocator_unlock(void)
 	  if (NEED_SUPRESSION_OF_MALLOC_AND_BFD)
 	  {
 	    channels::dc::malloc.off();
+#if CWDEBUG_LOCATION
 	    channels::dc::bfd.off();
+#endif
 	  }
 	  LibcwDoutScopeEnd;
 	}
@@ -1564,12 +1596,20 @@ void allocator_unlock(void)
 
       LIBCWD_DEFER_CANCEL;
       _private_::debug_channels.init(LIBCWD_TSD);
+
+      static _private_::debug_channels_ct hidden_channels;
+      hidden_channels.init(LIBCWD_TSD);
+
       DEBUG_CHANNELS_ACQUIRE_WRITE_LOCK;
 
       set_alloc_checking_off(LIBCWD_TSD);	// debug_channels is internal.
       _private_::debug_channels_ct::container_type& channels(_private_::debug_channels.write_locked());
       for(_private_::debug_channels_ct::container_type::iterator i(channels.begin()); i != channels.end(); ++i)
 	const_cast<char*>((*i)->get_label())[WST_max_len] = ' ';
+      _private_::debug_channels_ct::container_type& channels_not_listed(hidden_channels.write_locked());
+      for(_private_::debug_channels_ct::container_type::iterator i(channels_not_listed.begin());
+          i != channels_not_listed.end(); ++i)
+	const_cast<char*>((*i)->get_label())[WST_max_len] = ' ';
 
       // MT: This is not strict thread safe because it is possible that after threads are already
       //     running, a new shared library is dlopen-ed with a new global channel_ct with a larger
@@ -1587,6 +1627,9 @@ void allocator_unlock(void)
 
       for(_private_::debug_channels_ct::container_type::iterator i(channels.begin()); i != channels.end(); ++i)
 	const_cast<char*>((*i)->get_label())[WST_max_len] = '\0';
+      for(_private_::debug_channels_ct::container_type::iterator i(channels_not_listed.begin());
+          i != channels_not_listed.end(); ++i)
+	const_cast<char*>((*i)->get_label())[WST_max_len] = '\0';
       set_alloc_checking_on(LIBCWD_TSD);
 
 #if LIBCWD_THREAD_SAFE
@@ -1604,20 +1647,22 @@ void allocator_unlock(void)
       std::memset(WNS_label + label_len, ' ', max_label_len_c - label_len);
       WNS_label[WST_max_len] = '\0';
 
-      // We store debug channels in some organized order, so that the
-      // order in which they appear in the ForAllDebugChannels is not
-      // dependent on the order in which these global objects are
-      // initialized.
+      set_alloc_checking_off(LIBCWD_TSD);	// debug_channels is internal.
       if (add_to_channel_list)
       {
-	set_alloc_checking_off(LIBCWD_TSD);	// debug_channels is internal.
+	// We store debug channels in some organized order, so that the
+	// order in which they appear in the ForAllDebugChannels is not
+	// dependent on the order in which these global objects are
+	// initialized.
 	_private_::debug_channels_ct::container_type::iterator i(channels.begin());
 	for(; i != channels.end(); ++i)
 	  if (strncmp((*i)->get_label(), WNS_label, WST_max_len) > 0)
 	    break;
 	channels.insert(i, this);
-	set_alloc_checking_on(LIBCWD_TSD);
       }
+      else
+        channels_not_listed.push_back(this);
+      set_alloc_checking_on(LIBCWD_TSD);
 
       DEBUG_CHANNELS_RELEASE_WRITE_LOCK;
       LIBCWD_RESTORE_CANCEL;
