@@ -1,6 +1,6 @@
-// $Header: /cvsroot/libcwd/libcwd/include/libcwd/bfd.h,v 1.2 2003/05/22 10:56:58 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/include/libcwd/bfd.h,v 1.5 2004/07/15 10:14:34 libcw Exp $
 //
-// Copyright (C) 2000 - 2003, by
+// Copyright (C) 2000 - 2004, by
 // 
 // Carlo Wood, Run on IRC <carlo@alinoe.com>
 // RSA-1024 0x624ACAD5 1997-01-26                    Sign & Encrypt
@@ -36,9 +36,17 @@ extern "C" void* dlopen(char const*, int);
 extern "C" int dlclose(void*);
 #endif // LIBCWD_DLOPEN_DEFINED
 
-// Include the inline functions.
-#ifndef LIBCWD_CLASS_LOCATION_INL
-#include <libcwd/class_location.inl>
+#if CWDEBUG_ALLOC
+namespace libcwd {
+  namespace _private_ {
+    struct exit_function_list
+    {
+      struct exit_function_list* next;
+      // More here...
+    };
+    extern struct exit_function_list** __exit_funcs_ptr;
+  }
+}
 #endif
 
 #endif // LIBCWD_BFD_H
