@@ -1,6 +1,6 @@
-// $Header: /cvsroot/libcwd/libcwd/include/libcwd/private_threading.h,v 1.13 2004/05/27 03:03:53 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/include/libcwd/private_threading.h,v 1.15 2004/07/14 00:29:35 libcw Exp $
 //
-// Copyright (C) 2001 - 2003, by
+// Copyright (C) 2001 - 2004, by
 //
 // Carlo Wood, Run on IRC <carlo@alinoe.com>
 // RSA-1024 0x624ACAD5 1997-01-26                    Sign & Encrypt
@@ -109,7 +109,7 @@ extern unsigned int LIBCWD_DEBUGDEBUGLOCK_CERR_count;
 namespace libcwd {
 
 #if LIBCWD_DEBUGDEBUGRWLOCK
-__inline__
+inline
 _private_::raw_write_nt const&
 operator<<(_private_::raw_write_nt const& raw_write, pthread_mutex_t const& mutex)
 {
